from __future__ import absolute_import
import math
import numpy
import os
import sklearn.preprocessing
from ann_benchmarks.algorithms.base import BaseANN
from pymilvus import connections, Collection, FieldSchema, CollectionSchema, DataType, utility


class Milvus(BaseANN):

    def __init__(self, metric, index_type, params):
        self._metric = metric
        if metric == "euclidean":
            self._metric = "L2"
        elif metric == "angular":
            self._metric = "IP"
        self._index_type = index_type
        self._params = params
        self._collection_name = 'Milvus_%s' % (self._index_type)
        # connect to milvus
        milvus_host = os.environ.get("MILVUS_HOST", "localhost")
        milvus_port = os.environ.get("MILVUS_PORT", "19530")
        connections.connect(alias="default", host=milvus_host, port=milvus_port)

    def get_memory_usage(self):
        fd = os.popen('ssh root@${MILVUS_HOST} "ps -C milvus -o rss=,"')
        size_kb = int(fd.readline())
        return size_kb

    def fit(self, X):
        if self._metric == 'IP':
            X = sklearn.preprocessing.normalize(X, axis=1, norm='l2')

        # create collection
        fields = [
            FieldSchema("id", DataType.INT64, is_primary=True),
            FieldSchema("vector", dtype=DataType.FLOAT_VECTOR, dim=X.shape[1])
        ]
        self._collection = Collection(self._collection_name,
                                      schema=CollectionSchema(fields=fields,
                                                              description="%s" % self._params),
                                      shards_num=1)

        vector_ids = [id for id in range(len(X))]
        n_batches = math.ceil(X.nbytes / (1024 * 1024 * 100))
        batch_size = int(len(X) / n_batches)
        for i in range(0, len(X), batch_size):
            self._collection.insert([vector_ids[i:i+batch_size], X[i:i+batch_size]])

        # create index
        self._collection.create_index(
            "vector", {
                "metric_type": self._metric,
                "index_type": self._index_type,
                "params": self._params
            })
        print("index params:", self._collection.index().to_dict())
        self._collection.load()

    def set_query_arguments(self, params):
        self._search_params = params
        print("search params:", self._search_params)

    def query(self, v, n):
        if self._metric == 'IP':
            v /= numpy.linalg.norm(v)

        results = self._collection.search([v],
                                          anns_field="vector",
                                          limit=n,
                                          param={"metric_type": self._metric, "params": self._search_params})
        return results[0].ids

    def __str__(self):
        return 'Milvus(index_type=%s, params=%s)' % (
            self._index_type,
            self._params,
        )

    def done(self):
        # deleting collection
        utility.drop_collection(self._collection_name)
