from __future__ import absolute_import
import multiprocessing
import multiprocessing.pool
import os

from tair import Tair, tairvector
from ann_benchmarks.algorithms.base import BaseANN


class TairVector(BaseANN):
    def __init__(self, config, metric, index_type, method_param={}):
        try:
            import hiredis  # noqa
        except ImportError:
            raise ValueError("Install hiredis for better performance: `pip install hiredis`.")

        print("config:", config, "metric:", metric, "index_type:", index_type, "method_param:", method_param)
        self.is_binary = False
        self.index_type = index_type
        self.method_param = method_param
        if metric == "euclidean":
            self.metric = tairvector.DistanceMetric.L2
        elif metric == "angular":
            self.metric = "COSINE"
        elif metric == "jaccard":
            self.metric = tairvector.DistanceMetric.Jaccard
            self.is_binary = True
            self.method_param["data_type"] = tairvector.DataType.Binary
        else:
            # not supported
            raise ValueError("metric [%s] not supported" % metric)
        self.p = {}
        self.name = "TairVector(%s)" % str(self.method_param).replace("{", "[").replace("}", "]")

        tair_url = config.get("url", "redis://localhost:6379")
        self.client = Tair.from_url(tair_url)
        self.processes = int(config.get("parallelism", os.cpu_count()))

        # delete existing index
        self.client.tvs_del_index(self.name)

    def get_memory_usage(self):
        """Return the current memory usage of this algorithm instance
        (in kilobytes), or None if this information is not available."""
        # return in kB for backwards compatibility
        info = self.client.tvs_get_index(self.name)
        if info is None:
            return 0
        return int(info["index_data_size"]) / 1024

    def run_process(self, queue):
        while True:
            t = queue.get()
            if t is None:
                return
            start, batch = t
            p = self.client.pipeline(transaction=False)
            for i, v in enumerate(batch):
                # self.client.tvs_hset(self.name, str(start + i), v, self.is_binary)
                p.tvs_hset(self.name, str(start + i), v, self.is_binary)
            p.execute()

    def fit(self, X):
        ret = self.client.tvs_create_index(self.name, len(X[0]), self.metric, self.index_type, **self.method_param)

        # todo: handle failures
        if not ret:
            print("create index failed")

        queue = multiprocessing.Queue()
        batch_size = 32
        workers = [multiprocessing.Process(target=self.run_process, args=(queue,)) for _ in range(self.processes)]
        [worker.start() for worker in workers]

        for i in range(0, len(X), batch_size):
            queue.put((i, X[i : i + batch_size]))

        for _ in range(self.processes):
            queue.put(None)

        [worker.join() for worker in workers]

    def set_query_arguments(self, p):
        if self.index_type == "HNSW":
            self.p = {"ef_search": p}

    def query(self, v, n):
        result = self.client.tvs_knnsearch(self.name, n, v, self.is_binary, **self.p)
        return [int(s[0]) for s in result]

    def batch_query(self, X, n):
        pool = multiprocessing.pool.ThreadPool(processes=self.processes)
        self.res = pool.map(lambda q: self.query(q, n), X)

    def done(self):
        print("deleting index:", self.name)
        self.client.tvs_del_index(self.name)
